clc, clear
close all
tic
% select "my_UQLab_folder/core" as your working directory; see README.md
rng default
uqlab
%--- Labels
% output_lab='$${\tilde U_r^k} (m)$$';  
output_lab='$${\tilde U_d^k} (m)$$';
% output_lab='$${\tilde A_s^k} (m/s^2)$$';


var_labs={
    'm_c'        % x1: Main body mass [kg]
    'k_1'    % x2: Primary suspension stiffness [N/m]                 
    'k_2'    % x3: Secondary suspension stiffness [N/m] 
    'k_p'        % x4: Railpad vertical stiffness [N/m] 
    'm_s'        % x5: sleeper mass [kg]
    '\rho_{ba}'    % x6: ballast density (kg/m^3)
    'E_{ba}'       % x7: ballast elastic modulus (N/m^2) 
    'h_{ba}'       % x8: ballast height (m) 
    '\alpha_{ba}'         % x9: ballast load distribution angle (degree)
    'c_{ba}'       % x10: ballast damping (Ns/m) 
    'c_p'        % x11: Railpad vertical damping [Ns/m] 
    '\mu'       % x12: Mass per unit length [kg/m]
    'E'       % x13: Modulus of elasticity [N/m^2]
    'c'       % x14: Damping [%]
    'I'       % x15: Second moment of area [m^4]
    'I_{rr}'       % x16: irregularity amplitude
    };

%--- Definition of model variables: range

range=[
    22367.2471324521 44773.8153351932      
    946759.726879868 1505918.06324091
    339255.568798620 539620.639327993
    1e8 1.5e9
    220 325
    1500 2100
    8e7 1.6e8
    0.3 0.6
    15 35
    4e4 2.8e5  
    1e4 7e4    
    62579.2798676473 75420.7201323527
    28486225952.6856 41513774047.3144
    1.30209563778775 2.69790436221225
    45.3329177030852 57.2670822969148
    4.032e-7 10.8e-7];   

%--- Definition of the scenarios analysed 
% THRES_s=[1.5e-3 2e-3 2.5e-3]; % for Ur
THRES_s=[0.001,0.0015,0.058]; % for Ud
% THRES_s=[1.5*9.8,2*9.8,2.5*9.8]; % for As

MinMax='Min';
done=1;  %calculate optimization
nsim_v=10*[5 5 5 5 5 5 5 5 5 5 5 5 5 5 5 5]; %discretization of each variable

%% SAMPLING
nvars=length(nsim_v);
vars=cell(1,nvars);
seedtype=cell(1,nvars);
for v=1:nvars
    [vars{v},seedtype{v}]=Discretizing(var_labs{v},range(v,1),range(v,2),nsim_v(v),1e5);
end


%% CALCULATION OF EXTREME
if ~done
    t = cputime;
    nsimul=sum(nsim_v);
    Xs_vars=zeros(nsimul,nvars+1,length(THRES_s)); %[var1, var2, var3, ...,output] ^ [thr=1,2,3..]

    for thr=1:length(THRES_s)
        %--- ABSOLUTE OPTIMAL: 
        rng(1711);
        lb=range(:,1);ub=range(:,2);
        [OptimalPoint,OptimalValue,ValueAchieved]=optimizationGA(MinMax,THRES_s(thr),lb,ub,'');

        %--- OPTIMAL CURVE: one by one
        parOptFun=zeros(nvars,max(nsim_v)); %sol-function for each v,i
        parOptPoint=cell(nvars,max(nsim_v)); %sol-point for each v,i
        for v=1:nvars
            parfor i=1:nsim_v(v) 
                %--- Optimization: Min or Max value of the function at x0
                lb=range(:,1);ub=range(:,2);
                [lb(v,:),ub(v,:)]=deal(vars{v}(i));  %value of x0 is fixed
  
                [x_is,error2,ValueAchieved]=optimizationGA(MinMax,THRES_s(thr),lb,ub,'');

                fprintf('Ct=%-2.0f: Evaluation v=%-3.0f (i=%-2.0f) -> Error: %-8.2f\n',thr,v,i,error2^0.5)
                parOptFun(v,i)=ValueAchieved;
                parOptPoint{v,i}=x_is;
            end
        end
     
        pos=0;
        for v=1:nvars
            for i=1:nsim_v(v)
                pos=pos+1;
                Xs_vars(pos,nvars+1,thr)=parOptFun(v,i);
                Xs_vars(pos,1:nvars,thr)=parOptPoint{v,i};  
            end
        end
    end 
    
    comptime=cputime-t; fprintf('CPU time (seconds): %f\n',comptime)
    save('SA_thres_v3_Ud.mat','Xs_vars','OptimalValue','OptimalPoint','comptime')
else 
    load('SA_thres_v3_Ud.mat','Xs_vars','OptimalValue','OptimalPoint','comptime')
end 

%% RESULTS

%--- Computing local sensitivity 
ExtSens=cell(nvars,length(THRES_s));  %one column for each threshold considered
for thr=1:length(THRES_s)
    pos=0;
    for v=1:nvars           
       output=Xs_vars(pos+1:pos+nsim_v(v),nvars+1,thr);
       ExtSens{v,thr}=SensComp(output,vars{v});    
       pos=pos+nsim_v(v);
    end
end

%--- Plot curves
figure,nsubplots=nvars; 
sel_var0=1:nvars; 
%bounds for ploting
selTHs=1:3;
PlotBounds=[min(min(Xs_vars(:,nvars+1,selTHs))),max(max(Xs_vars(:,nvars+1,selTHs)))];
col=['r','b','g','k','m'];
pos=0;
for v=1:nvars
    pos=pos+nsim_v(v);
    for thr=selTHs
        output=Xs_vars(pos-nsim_v(v)+1:pos,nvars+1,thr);
        pos_plot=find(sel_var0==v); 
        subplot(ceil(nsubplots/3),3,pos_plot)
        SensCurvePlot(output,vars{v},var_labs{v},output_lab,seedtype{v},[],THRES_s(thr),PlotBounds,col(thr),1);
    end
end
pos = get(gcf, 'Position');
set(gcf, 'Position',pos+[0 -500 500 1500]) 
exportgraphics(gcf,'Ud_thres_v3.eps','BackgroundColor','none')

toc

function [x_is,z,ValueAchieved]=optimizationGA(MinMax,THRS,lb,ub,x_is0)
A_eq=[]; b_eq=[];

A_in=[]; b_in=[];

nvars=length(lb);

if isempty(x_is0)
    options = optimoptions('ga','Display','off');
else
    options = optimoptions('ga','MigrationDirection','both','InitialPopulationMatrix',x_is0','Display','off');

end

[x_is,z,exitflag]=ga(@equation,nvars,A_in,b_in,A_eq,b_eq,lb,ub,[],[],options); 
ValueAchieved=PCE(x_is');

if strcmp(MinMax,'Max'), z=-z; end

%--- Objective function
    function y=equation(x_is)
        Ypce=PCE(x_is');
        y=(Ypce-THRS)^2;
        if strcmp(MinMax,'Max'), y=-y; end
    end

    function Ur=PCE(x_is) 
        % replace Ur here and in the following by other design labels
        persistent myModel_Ur_v3
        if isempty(myModel_Ur_v3 )
        load('myPCE_Ur_v3.mat','myModel_Ur_v3')
        end     
        Ur = uq_evalModel(myModel_Ur_v3 ,x_is');
    end

end

function E_j=SensComp(matrix,ticksX)
%--- Computing sensitivities
dX_j=ticksX(2:end)-ticksX(1:end-1);

% Local sensitivity: how a variation of the var @x0 affects the extremesystem response
dY_j=matrix(2:end)-matrix(1:end-1);
E_j=dY_j'./dX_j;
end
 
function SensCurvePlot(matrix,ticksX,labelX,labelY,seedtype,disc_var,Ct,PlotBounds,col,plotoption)
%---Plotting maximun or minimum
if plotoption
    hold on
    %
    if strcmp('log',seedtype)
        plot(matrix,['-',col],'Linewidth',2)
        % horizontal line
        yline(Ct,['--',col],'LineWidth',1)
        labelX=[labelX, ' (log scale)'];
        pos=floor(linspace(1,length(ticksX),4));
        xlab=(vector2str_cell(ticksX,2,pos));
        set(gca,'xtick',1:length(ticksX),'xticklabel',xlab);

    elseif disc_var
        scatter(ticksX,matrix,col,'filled')
        plot(ticksX,matrix,[':',col],'Linewidth',1)
        % horizontal line
        yline(Ct,['--',col],'LineWidth',1)
        xlim([ticksX(1), ticksX(end)])

    else
        plot(ticksX,matrix,['-',col],'Linewidth',2)
        % horizontal line
        yline(Ct,['--',col],'LineWidth',1)
        xlim([ticksX(1), ticksX(end)])
    end
    xlabel(labelX)
    ylim(PlotBounds)
    ylabel(labelY,'Interpreter','latex')
    box on, grid on
end
end

function s=vector2str_cell(vector,precision,pos)
%only the values at "pos" are written
n=length(vector);
s=cell(n,1);
for i=1:n
    if any(i==pos)
        s{i}=mat2str(vector(i),precision);
    else
        s{i}='';
    end
end
end

function [dist,typeS]=Discretizing(varname,range1, range2, dist_length,threshold)
if (range2-range1)>threshold
    dist=logspace(log10(range1),log10(range2),dist_length);
    typeS='log';
    fprintf('%s: Log\n',varname)

else
    dist=linspace(range1,range2,dist_length);
    typeS='lin';
    fprintf('%s: Lin\n',varname)
end
end